import re
from textrank4zh import TextRank4Keyword
# from .rake import *
# import pandas as pd
import numpy as np
# from .util import PDFProcessor
# from scipy.spatial.distance import cosine
from bert_serving.client import BertClient
# import math
# import jieba
import requests
import json
# from difflib import SequenceMatcher 


class TextRank:
    def __init__(self, n_word, word_min_len, sentence_min_len):
        self.n_word = n_word
        self.word_min_len = word_min_len
        self.sentence_min_len = sentence_min_len

    def get_keywords(self, text):
        tr4w = TextRank4Keyword()
        tr4w.analyze(text=text, lower=True, window=2) 
        kw_list = []
        for item in tr4w.get_keywords(self.n_word, word_min_len=self.word_min_len):
            kw_list.append(item.word)
        return kw_list

    def get_key_sentences(self, kw_list, text):
        text_list = re.split(',|。',text)
        sentence_list = list(filter(lambda x: len(x) > self.sentence_min_len, text_list))
        sentence_set = set(sentence_list)
        sentence_dict = {s:0 for s in list(sentence_set)}
        for word in kw_list:
            for sentence, occ in sentence_dict.items():
                if word in sentence:
                    sentence_dict[sentence] += 1
        important_sentece = {k:v for k, v in sentence_dict.items() if v != 0}
        return important_sentece

class PretrainedBert:
    def __init__(self, api_url, api_port):
        self.model = BertClient(ip=api_url, port=api_port)
    
    def encode(self, text_list):
        embeddings = np.zeros((len(text_list), 768))
        if '' in text_list:
            for index, text in enumerate(text_list):
                if text == '':
                    continue
                else:
                    embeddings[index] = self.model.encode([text])[0]
            return embeddings
        else:
            embeddings = self.model.encode(text_list)
            return embeddings

class GPT2ForQuestionGeneration:
    def __init__(self, api_url, api_port):
        self.api_url = 'http://' + api_url + ':' + str(api_port) + '/generate'
    
    def generate(self, sentences):
        data = json.loads(json.dumps({'sentences': sentences}))
        # print(data)
        resp = requests.get(self.api_url, json=data)
        # print(resp.text)
        questions = dict(resp.json())['quesiotns']
        return questions

# class Rake:
#     def __init__(self, n_word, sentence_min_len):
#         self.n_word = n_word
#         self.sentence_min_len = sentence_min_len

#     def get_keywords(self, text):
#         kw_tuple = run(text)
#         kw_list = [word for (word, weight) in kw_tuple]
#         return kw_list[:self.n_word]

#     def get_key_sentences(self, kw_list, text):
#         text_list = re.split(',|。',text)
#         sentence_list = list(filter(lambda x: len(x) > self.sentence_min_len, text_list))
#         sentence_set = set(sentence_list)
#         sentence_dict = {s:0 for s in list(sentence_set)}
#         for word in kw_list:
#             for sentence, occ in sentence_dict.items():
#                 if word in sentence:
#                     sentence_dict[sentence] += 1
#         important_sentece = {k:v for k, v in sentence_dict.items() if v != 0}
#         return important_sentece

# class BertSum:
#     def __init__(self, ratio, article):
#         self.model = BertClient(ip="140.115.53.158", port=5555)
#         self.ratio = ratio
#         self.article = article
#         self.concept = self.model.encode(list(self.article))[0]
#         sentence_cnt = 0
#         page_content = article.split('。')
#         if '' in page_content:
#             page_content.remove('')
#         for page in page_content:
#             sentences = page.split(',')
#             if '' in sentences:
#                 sentences.remove('')
#             for s in sentences:
#                 sentence_cnt += 1
#         self.num_sentence = math.ceil(sentence_cnt*self.ratio)
    
#     def get_key_sentences(self):
#         distance = {}
#         page_content = self.article.split('。')
#         if '' in page_content:page_content.remove('')
#         for page in page_content:
#             sentences = page.split(',')
#             if '' in sentences:sentences.remove('')
#             for s in sentences:
#                 embedding = self.model.encode(list(s))[0]
#                 distance.update({s: self.get_distance(self.concept, embedding)})
#         sort_dict = {k: v for k, v in sorted(distance.items(), key=lambda item: item[1])}
#         dict_slice = lambda adict, start, end: { k:adict[k] for k in list(adict.keys())[start:end]}
#         return dict_slice(sort_dict, 0, self.num_sentence)

#     def get_distance(self, p1, p2):
#         return cosine(p1, p2)

# class BertForEmbedding:
#     def __init__(self):
#         self.model = BertClient(ip="140.115.53.158", port=5555)
    
#     def embed_slide(self, slide_path):
#         pdf = PDFProcessor(slide_path)
#         page_centrals = []
#         for i in range(pdf.pageCount):
#             text_list = pdf.get_page_text(i+1)
#             text = '。'.join(text_list)
#             if text != '':
#                 page_central = self.model.encode([text])
#             else:
#                 page_central = np.zeros((1, 768))
#             page_centrals.append(page_central)
#         return page_centrals
#     def embed_text(self, text):
#         if text != '':
#             embedding = self.model.encode([text])
#         else:
#             embedding = np.zeros((1, 768))
#         return embedding

# class MemoFeaturesExtractor:
#     def __init__(self, content_embedding, slide_raw_text, slide_kw_list):
#         self.content_embedding = content_embedding
#         self.slide_raw_text = slide_raw_text
#         self.slide_kw_list = slide_kw_list
#         # jieba.enable_paddle()

#     def get_length(self, memo_text):
#         memo_text = memo_text.replace(";:::sgquotation:::;", "'").replace(";:::dbquotation:::;", "\'\'").replace(";:::nl:::;", "\n")
#         # tokenizer = hanlp.utils.rules.tokenize_english
#         # length = len(tokenizer(memo_text))
#         # if length <= 1:
#         #     #處理中文
#         return len(memo_text)
#     def get_lcs(self, memo_text):
#         memo_text = memo_text.replace(";:::sgquotation:::;", "'").replace(";:::dbquotation:::;", "\'\'").replace(";:::nl:::;", "\n")
#         seqMatch = SequenceMatcher(None, self.slide_raw_text, memo_text)
#         match = seqMatch.find_longest_match(0, len(self.slide_raw_text), 0, len(memo_text))

#         return match.size
#     def get_semantic_similarity(self, memo, page_no):
#         model = BertForEmbedding()
#         memo_embedding = model.embed_text(memo)[0]
#         # spec_page_text = self.slide_page_text[page_no - 1]
#         # url = 'http://140.115.53.158:5000/embedding' #call API
#         # model = BertForEmbedding(model_path='../model')
#         page_central = self.content_embedding[page_no - 1]
#         cos_sim = 1 - cosine(page_central[0], memo_embedding) / 2
        
#         return cos_sim
#     def get_kw_ratio(self, memo_text):
#         memo_text = memo_text.replace(";:::sgquotation:::;", "'").replace(";:::dbquotation:::;", "\'\'").replace(";:::nl:::;", "\n")
#         tokens = jieba.cut(memo_text)
#         hit_kw = list(set(tokens).intersection(set(self.slide_kw_list)))
#         return 0.8 + len(hit_kw) / len(self.slide_kw_list)

# class BertForGuiding:
#     def __init__(self, model_path):
#         # self.use_cuda = use_cuda

#         self.model = BertClient(ip="140.115.53.158", port=5555)
#         # if self.use_cuda:
#         #     device = torch.device('cuda')
#         #     self.model.to(device)
#         # else:
#         #     device = torch.device('cpu')
#         #     self.model.to(device)

#     def get_central(self, pdf):
#         page_central = []
#         for i in range(pdf.pageCount):
#             text = ','.join(pdf.get_page_text(i+1))
# #     for token in pdf.get_page_text(i + 1):
#             feature = self.model.encode([text])
#             page_central.append(feature)
#         return page_central
        
#     def get_ans_feature(self, ans):
#         ans_feature = self.model.encode([ans])
#         return ans_feature

#     def get_distance(self, ans_feature, centrals):
#         index = 0
#         distance_array = []
#         for pc in centrals:
#             distance_array.append((index + 1, cosine(pc.numpy(), ans_feature.numpy())))
#         return distance_array



from .model import PretrainedBert
from .util import PDFProcessor
from scipy.spatial.distance import cosine
from operator import itemgetter
import numpy as np

class Recommender(object):
    def __init__(self, pdf_path, num_page, api_url, api_port):
        pdf = PDFProcessor(file_path=pdf_path)
        pages_content = []
        for page in range(pdf.pageCount):
            pages_content.append(','.join(pdf.get_page_text(page + 1)))
        self.pages_content = pages_content
        self.bert_model = PretrainedBert(api_url=api_url, api_port=api_port)
        self.pages_embedding = self.bert_model.encode(self.pages_content)
        self.num_page = num_page

    def guiding_from(self, ta_ans):
        concept = self.bert_model.encode([ta_ans])[0]
        page_distence = []
        for page_no, embedding in enumerate(self.pages_embedding):
            if np.count_nonzero(embedding) == 0:
                distance = np.inf
            else:
                distance = cosine(embedding, concept)
            page_distence.append((page_no, distance))
        page_distence = sorted(page_distence, key=itemgetter(1))
        return page_distence[:self.num_page]



from .model import TextRank, GPT2ForQuestionGeneration
from .util import SentencePreprocessor, PDFProcessor
from tqdm import tqdm

class Generator(object):
    def __init__(self, pdf_path, translate_api_url, translate_api_port, gpt2_api_url, gpt2_api_port):
        self.pdf = PDFProcessor(pdf_path)
        self.text = self.pdf.get_pdf_text()
        self.sumarization_model = TextRank(n_word=5, word_min_len=2, sentence_min_len=6)
        self.preprocessor = SentencePreprocessor(api_url=translate_api_url, api_port=translate_api_port)
        self.gpt2 = GPT2ForQuestionGeneration(api_url=gpt2_api_url, api_port=gpt2_api_port)

    def summarize(self):
        print('=====Start Summarization=====')
        kw_list = self.sumarization_model.get_keywords(self.text)
        important_sentece = self.sumarization_model.get_key_sentences(kw_list, self.text)
        print('=====Summarization finished=====')
        return list(important_sentece.keys())
    
    def filt(self, sentence_list):
        regular_sentence = []
        print('=====Start filter=====')
        pbar = tqdm(sentence_list)
        for index, sentence in enumerate(pbar):
            pbar.set_description("Processing sentence %s" % index)
            if self.preprocessor.is_sentence(sentence):
                regular_sentence.append(sentence)
            else:
                pass
        print('=====Filter finished=====')
        return regular_sentence
    
    def translate(self, regular_sentence_list):
        eng_sentence = []
        print('=====Start Translation=====')
        pbar = tqdm(regular_sentence_list)
        for index, sentence in enumerate(pbar):
            pbar.set_description("Processing sentence %s" % index)
            tgt_sentence = self.preprocessor.translateToEnglish(sentence)
            eng_sentence.append(tgt_sentence)
        print('=====Translation finished=====')
        return eng_sentence

    def generate_guestion(self, regular_sentence_list):
        qa_list = []
        print('=====Start generation=====')
        eng_sentences = self.translate(regular_sentence_list)
        questions = self.gpt2.generate(eng_sentences)
        for q, a in zip(questions, regular_sentence_list):
            qa_list.append((q, a))
        print('=====Generation finished=====')
        return qa_list
    
    def get_qa(self):
        important_senteces = self.summarize()
        regular_sentences = self.filt(important_senteces)
        qas = self.generate_guestion(regular_sentences)
        return qas


from .model import TextRank, PretrainedBert
from .util import PDFProcessor
from nltk.translate.bleu_score import sentence_bleu, SmoothingFunction
from sklearn.metrics.pairwise import cosine_similarity
from bert_serving.client import BertClient
from difflib import SequenceMatcher 
import jieba

class Grader(object):
    def __init__(self, pdf_path, bert_api_url, bert_api_port, n_word=5, word_min_len=2, sentence_min_len=6):
        self.pdf = PDFProcessor(pdf_path)
        self.text = self.pdf.get_pdf_text()

        self.summarize_model = TextRank(n_word=n_word, word_min_len=word_min_len, sentence_min_len=sentence_min_len)
        self.bert_model = PretrainedBert(api_url=bert_api_url, api_port=bert_api_port)
        self.slides_embedding = self.bert_model.encode([self.text])
        
        self.keywords = self.summarize_model.get_keywords(self.text)
        self.key_sentences = self.summarize_model.get_key_sentences(self.keywords, self.text)
    
    def grade_marker(self, marker_text):
        chencherry = SmoothingFunction()
        reference = self.key_sentences
        candidate = marker_text
        bleu = sentence_bleu(reference, candidate, weights=(1, 0, 0, 0), smoothing_function=chencherry.method1)
        return bleu

    def grade_memo(self, memo_text):
        #get length
        length = len(memo_text) 

        #get longest common sequence
        seqMatch = SequenceMatcher(None, self.text, memo_text)
        match = seqMatch.find_longest_match(0, len(self.text), 0, len(memo_text))
        lcs = match.size 

        #get semantic similarity
        memo_embedding = self.bert_model.encode([memo_text])
        semantic_similarity = cosine_similarity(memo_embedding, self.slides_embedding)
        if semantic_similarity <= 0:
            semantic_similarity = 0.1
        else:
            semantic_similarity = semantic_similarity[0][0]

        #get keyword hit ratio
        tokens = jieba.cut(memo_text)
        hit_kw = list(set(tokens).intersection(set(self.keywords)))
        kw_hit_ratio = 0.8 + len(hit_kw) / len(self.keywords)

        memo_score = (length - lcs)*semantic_similarity*kw_hit_ratio
        return memo_score


# class ShortAnswerGrader(object):


import fitz
import os
import numpy as np
from ckippy import parse_tree
import requests


class PDFProcessor:
    def __init__(self, file_path):
        self.path = file_path
        self.doc = fitz.open(self.path)
        self.pageCount = self.doc.pageCount
        # self.text = ''
    def is_chinese(self, uchar):
        if uchar >= u'\u4e00' and uchar<=u'\u9fa5':
            return True
        else:
            return False
        
    def is_number(self, uchar):
        if uchar >= u'\u0030' and uchar<=u'\u0039':
            return True
        else:
            return False
        
    def is_alphabet(self, uchar):
        if (uchar >= u'\u0041' and uchar<=u'\u005a') or (uchar >= u'\u0061' and uchar<=u'\u007a'):
            return True
        else:
            return False
    def is_char(self, uchar):
        if not (self.is_chinese(uchar) or self.is_number(uchar) or self.is_alphabet(uchar)):
            return True
        else:
            return False
    def get_pdf_text(self):
        page_content = []
        all_page = []

        for page in self.doc:
            page_content.append(page.getText('text'))
        for page_num in range(len(page_content)):
            for j in page_content[page_num]:
                if self.is_char(j):
                    page_content[page_num] = page_content[page_num].replace(j, ',')
        for page_num in range(len(page_content)):
            page_content[page_num] = list(filter(lambda x: len(x) > 1, page_content[page_num].split(',')))

        for page in page_content:
            all_page.append(','.join(page))    
        text = '。'.join(all_page)
        return text

    def get_page_text(self, page):
        page_content = []
        # all_page = []

        for p in self.doc:
            page_content.append(p.getText('text'))
        for page_num in range(len(page_content)):
            for j in page_content[page_num]:
                if self.is_char(j):
                    page_content[page_num] = page_content[page_num].replace(j, ',')
        for page_num in range(len(page_content)):
            page_content[page_num] = list(filter(lambda x: len(x) > 1, page_content[page_num].split(',')))
        
        return page_content[page-1]
    
    def highlight(self, important_sentence):
        # doc = fitz.open(self.path)
        slide_name = self.path.split('/')[-1].replace('.pdf', '')
        for page in self.doc:
            for key_sentence, weight in important_sentence.items():
                quads = page.searchFor(key_sentence, hit_max=100, quads=True)
                page.addHighlightAnnot(quads)
        file_name = slide_name + 'marked' + '.pdf'
        store_path = os.path.join('./output_pdf', file_name)
        self.doc.save(store_path)

class SentencePreprocessor:
    def __init__(self, api_url, api_port):
        self.api_url = 'http://' + api_url + ':' + str(api_port) + '/translate'

    def is_sentence(self, text):
        tree = parse_tree(text)
        result = tree[0]
        if 'S' in result.split('(')[0] and len(text) >= 10:
            return True
        else:
            return False

    def translateToChinese(self, src_text):
        params = {'text': src_text, 'src': 'en', 'tgt': 'zh-TW'}
        resp = requests.get(self.api_url, params=params)
        return dict(resp.json())['translatedText']

    def translateToEnglish(self, src_text):
        params = {'text': src_text, 'src': 'zh-TW', 'tgt': 'en'}
        resp = requests.get(self.api_url, params=params)
        return dict(resp.json())['translatedText']




import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="autota", # 
    version="0.0.3",
    author="Jay",
    author_email="a121406@gmail.com",
    description="Automatic generate QA from slides and grade marker/memo.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/miyuiki/autota",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)